/* **********************************************************
 * Copyright (C) 1998-2000 VMware, Inc.
 * All Rights Reserved
 * **********************************************************/

#include "driver-config.h"

#ifdef KERNEL_2_1
#  define EXPORT_SYMTAB
#  ifndef __KERNEL__
#    define __KERNEL__
#  endif
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/sched.h>
#ifdef KERNEL_2_2
#  include <linux/slab.h>
#else
#  include <linux/malloc.h>
#endif
#ifdef KERNEL_2_1
#include <linux/poll.h>
#include <asm/uaccess.h>
#endif

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/mm.h>
#include <linux/skbuff.h>
#include <linux/if_ether.h>
#include <linux/sockios.h>
#include <net/sock.h>

#define __KERNEL_SYSCALLS__
#include <asm/io.h>

#include <linux/proc_fs.h>
#include <linux/file.h>

#include "vm_oui.h"
#include "vnet.h"
#include "vnetInt.h"

#include "vmnetInt.h"

typedef struct VNetUserIFStats {
   unsigned    read;
   unsigned    written;
   unsigned    queued;
   unsigned    droppedDown;
   unsigned    droppedMismatch;
   unsigned    droppedOverflow;
} VNetUserIFStats;

typedef struct VNetUserIF {
   VNetPort               port;
   struct sk_buff_head    packetQueue;
   uint32*                pollPtr;
   struct page*           pollPage;
   uint32                 pollMask;
   uint32                 clusterCount;
#ifndef KERNEL_2_3_1
   struct wait_queue     *waitQueue;
#else
   wait_queue_head_t      waitQueue;
#endif
   VNetUserIFStats        stats;
} VNetUserIF;


static INLINE int VNetUserIfMapPollPtr(VNetUserIF *userIf, uint32 userVA);
static INLINE void VNetUserIfUnmapPollPtr(VNetUserIF *userIf);

static void VNetUserIfFree(VNetJack *this);
static void VNetUserIfReceive(VNetJack *this, struct sk_buff *skb);

static int  VNetUserIfRead(VNetPort *port, struct file *filp,
                           char *buf, size_t count);
static int  VNetUserIfWrite(VNetPort *port, struct file *filp,
                            const char *buf, size_t count);
static int  VNetUserIfIoctl(VNetPort *this, struct file *filp,
                            unsigned int iocmd, unsigned long ioarg);   

#ifdef KERNEL_2_1
static int  VNetUserIfPoll(VNetPort *this, struct file *filp,
                           poll_table *wait);
#else
static int  VNetUserIfSelect(VNetPort *this, struct file *filp,
                             int sel_type, select_table *wait);
#endif

#ifdef KERNEL_2_1
static int VNetUserIfProcRead(char *page, char **start, off_t off,
                              int count, int *eof, void *data);
#endif

/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfMapPollPtr --
 *
 *      Creates permanent mapping for poll structure provided by user.
 * 
 * Results: 
 *      Returns zero on OK, -EINVAL on failure.
 *
 * Side effects:
 *      Fields pollPtr and pollPage are filled in VNetUserIf structure.
 *
 * Bugs:
 *      Polling is disabled by BIGMEM patches.
 *
 *----------------------------------------------------------------------
 */

static INLINE int
VNetUserIfMapPollPtr(VNetUserIF *userIf, uint32 userVA)
{
#if defined(KERNEL_2_1)
   pte_t* pte;
   char* ptr;
   unsigned long physPage;
   char dummy;

   if (userIf->pollPtr) {
      printk(KERN_DEBUG "vmnet: Polling already active\n");
      return -EBUSY;
   }
   /* poll must not cross page boundary */
   if (((userVA + sizeof(uint32) - 1) & (PAGE_SIZE - 1)) < 
       sizeof(uint32) - 1) {
      return -EINVAL;
   }
   memcpy_fromfs(&dummy,(void *)userVA, 1);
   memcpy_tofs((void *)userVA, &dummy, 1);
   pte = pte_offset_map(pmd_offset(pgd_offset(current->mm, userVA),
			       userVA), userVA);
   physPage = PTE_2_PFN(pte_val(*pte));
   pte_unmap(pte);
   get_page(mem_map + physPage);
   userIf->pollPage = mem_map + physPage;
#ifdef KERNEL_2_3_25
   ptr = (void *) kmap(mem_map + physPage);
#else
#ifdef CONFIG_BIGMEM
#define MAXMEM (unsigned long)(-__PAGE_OFFSET-(64 << 20))
   if ((physPage << PAGE_SHIFT) >= MAXMEM) {
      return -EINVAL;
   }
#endif
   ptr = (void *)__PAGE_OFFSET + (physPage << PAGE_SHIFT);
#endif
   ptr += (userVA & (PAGE_SIZE-1));
   userIf->pollPtr = (uint32*)ptr;
   return 0;
#endif
   return -EINVAL;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfUnmapPollPtr --
 *
 *      Destroys permanent mapping for poll structure provided by user.
 * 
 * Results: 
 *      None.
 *
 * Side effects:
 *      Fields pollPtr, pollPage and pollMask in VNetUserIf structure
 *      are cleared.
 *
 *----------------------------------------------------------------------
 */

static INLINE void
VNetUserIfUnmapPollPtr(VNetUserIF *userIf)
{
#ifdef KERNEL_2_1
   if (userIf->pollPage) {
#ifdef KERNEL_2_3_25
      kunmap(userIf->pollPage);
#endif
      put_page(userIf->pollPage);
   } else {
      printk(KERN_DEBUG "vmnet: Polling was already deactivated\n");
   }
#endif
   userIf->pollPtr = NULL;
   userIf->pollPage = NULL;
   userIf->pollMask = 0;
}

/*
 *----------------------------------------------------------------------
 *
 * VNetUserIf_Create --
 *
 *      Create a user level port to the wonderful world of virtual
 *      networking.
 * 
 * Results: 
 *      Errno. Also returns an allocated port to connect to,
 *      NULL on error.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetUserIf_Create(VNetPort **ret)
{
   VNetUserIF *userIf = NULL;
   static unsigned id = 0;
   int retval;
   
   userIf = (struct VNetUserIF*)kmalloc(sizeof(struct VNetUserIF), GFP_KERNEL);
   if (!userIf) {
      return -ENOMEM;
   }

   /*
    * Initialize fields.
    */
   
   userIf->port.id = id++;

   userIf->port.jack.peer = NULL;
   userIf->port.jack.numPorts = 1;
   sprintf(userIf->port.jack.name, "userif%d", userIf->port.id);
   userIf->port.jack.private = userIf;
   userIf->port.jack.index = 0;
   userIf->port.jack.procEntry = NULL;
   userIf->port.jack.free = VNetUserIfFree;
   userIf->port.jack.rcv = VNetUserIfReceive;
   userIf->port.jack.cycleDetect = NULL;
   userIf->port.jack.portsChanged = NULL;
   userIf->port.jack.isBridged = NULL;
   userIf->pollPtr = NULL;
   userIf->clusterCount = 0;
   userIf->pollMask = 0;

   /*
    * Make proc entry for this jack.
    */
   
   retval = VNetProc_MakeEntry(NULL, userIf->port.jack.name, S_IFREG,
                               &userIf->port.jack.procEntry);
   if (retval) {
      if (retval == -ENXIO) {
         userIf->port.jack.procEntry = NULL;
      } else {
         kfree(userIf);
         return retval;
      }
   } else {
#ifdef KERNEL_2_1
      userIf->port.jack.procEntry->read_proc = VNetUserIfProcRead;
      userIf->port.jack.procEntry->data = userIf;
#else
      userIf->port.jack.procEntry = NULL;
#endif
   }

   /*
    * Rest of fields.
    */
   
   userIf->port.flags = IFF_RUNNING;

   memset(userIf->port.paddr, 0, sizeof userIf->port.paddr);
   memset(userIf->port.ladrf, 0, sizeof userIf->port.ladrf);

   userIf->port.paddr[0] = VMX86_OUI0;
   userIf->port.paddr[1] = VMX86_OUI1;
   userIf->port.paddr[2] = VMX86_OUI2;

   userIf->port.fileOpRead = VNetUserIfRead;
   userIf->port.fileOpWrite = VNetUserIfWrite;
   userIf->port.fileOpIoctl = VNetUserIfIoctl;
#ifdef KERNEL_2_1
   userIf->port.fileOpPoll = VNetUserIfPoll;
#else
   userIf->port.fileOpSelect = VNetUserIfSelect;
#endif
   
   skb_queue_head_init(&(userIf->packetQueue));
#ifndef KERNEL_2_3_1
   userIf->waitQueue = NULL;
#else
   init_waitqueue_head(&userIf->waitQueue);
#endif

   memset(&userIf->stats, 0, sizeof(VNetUserIFStats));
   
   *ret = (VNetPort*)userIf;
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfFree --
 *
 *      Free the user interface port.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

void
VNetUserIfFree(VNetJack *this)
{
   VNetUserIF *userIf = (VNetUserIF*)this;
   struct sk_buff *skb;

   for (;;) {
      skb = skb_dequeue(&userIf->packetQueue);
      if (skb == NULL) {
	 break;
      }
      DEV_KFREE_SKB(skb, FREE_WRITE);
   }
   
   if (userIf->pollPtr) {
      VNetUserIfUnmapPollPtr(userIf);
      userIf->clusterCount = 0;
   }

   if (this->procEntry) {
      VNetProc_RemoveEntry(this->procEntry, NULL);
   }

   kfree(userIf);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfReceive --
 *
 *      This jack is receiving a packet. Take appropriate action.
 *
 * Results: 
 *      None.
 *
 * Side effects:
 *      Frees skb.
 *
 *----------------------------------------------------------------------
 */

void
VNetUserIfReceive(VNetJack        *this,
                  struct sk_buff  *skb)
{
   VNetUserIF *userIf = (VNetUserIF*)this->private;
   uint8 *dest = SKB_2_DESTMAC(skb);
   
   if (!UP_AND_RUNNING(userIf->port.flags)) {
      userIf->stats.droppedDown++;
      goto drop_packet;
   }
   
   if (!VNetPacketMatch(dest,
                        userIf->port.paddr,
                        userIf->port.ladrf,
                        userIf->port.flags)) {
      userIf->stats.droppedMismatch++;
      goto drop_packet;
   }
   
   if (skb_queue_len(&userIf->packetQueue) >= VNET_MAX_QLEN) {
      userIf->stats.droppedOverflow++;
      goto drop_packet;
   }
   
   userIf->stats.queued++;

#ifndef KERNEL_2_1
   start_bh_atomic();
#endif
   skb_queue_tail(&userIf->packetQueue, skb);
   if (userIf->pollPtr && 
        (skb_queue_len(&userIf->packetQueue) >= userIf->clusterCount)) {
      *userIf->pollPtr |= userIf->pollMask;
   }
#ifndef KERNEL_2_1
   end_bh_atomic();
#endif
   wake_up(&userIf->waitQueue);
   return;
   
 drop_packet:
   DEV_KFREE_SKB(skb, FREE_WRITE);
}


/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfRead --
 *
 *      The virtual network's read file operation. Reads the next pending
 *      packet for this network connection.
 *
 * Results: 
 *      On success the len of the packet received,
 *      else if no packet waiting and nonblocking 0,
 *      else -errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int 
VNetUserIfRead(VNetPort     *port,
               struct file  *filp,
               char         *buf,
               size_t        count)
{
   VNetUserIF *userIf = (VNetUserIF*)port->jack.private;
   struct sk_buff *skb;
#ifdef KERNEL_2_1
   int ret;

#ifdef HACK_DECLARE_WAITQUEUE
   /*
    * This hack was added because of the problem we had with gcc 2.7.2.3
    * calling memset() when initializing DECLARE_WAITQUEUE()
    * This test is to detect if the kernel header linux/wait.h 
    * struct wait_queue changes in size.
    */

   struct wait_queue wait;
   wait.task = current;
   wait.next = NULL;
   wait.flags = 0;
#else
   DECLARE_WAITQUEUE(wait, current);
#endif
   
#endif

#ifdef KERNEL_2_1
   add_wait_queue(&userIf->waitQueue, &wait);
   for (;;) {
      current->state = TASK_INTERRUPTIBLE;
      mb();
      ret = -EAGAIN;
      skb = skb_dequeue(&userIf->packetQueue);
      if (userIf->pollPtr && skb_queue_empty(&userIf->packetQueue)) {
         *userIf->pollPtr &= ~userIf->pollMask;
      }  
      if (skb != NULL || filp->f_flags & O_NONBLOCK) {
         break;
      }
      ret = -EINTR;
      if (signal_pending(current)) {
         break;
      }
      schedule();
   }
   current->state = TASK_RUNNING;
   remove_wait_queue(&userIf->waitQueue, &wait);
   if (! skb) {
      return ret;
   }
#else
   for (;;) {
      start_bh_atomic();
      skb = skb_dequeue(&userIf->packetQueue);
      if (userIf->pollPtr && skb_queue_empty(&userIf->packetQueue)) {
         *userIf->pollPtr &= ~userIf->pollMask;
      }  
      end_bh_atomic();
      if (skb != NULL || filp->f_flags & O_NONBLOCK) {
	 break;
      }
      interruptible_sleep_on(&userIf->waitQueue);
      if (current->signal & ~current->blocked) {
	 return -EINTR;
      }
   }

   if (! skb) {
      return -EAGAIN;
   }
#endif

   userIf->stats.read++;

   count = (count < skb->len) ? count : skb->len;
   memcpy_tofs(buf, skb->data, count);

   DEV_KFREE_SKB(skb, FREE_WRITE);
   return count;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfWrite --
 *
 *      The virtual network's write file operation. Send the raw packet
 *      to the network.
 *
 * Results: 
 *      On success the count of bytes written else errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int 
VNetUserIfWrite(VNetPort     *port,
                struct file  *filp,
                const char   *buf,
                size_t        count)
{
   VNetUserIF *userIf = (VNetUserIF*)port->jack.private;
   struct sk_buff *skb;

   /*
    * Check size
    */
   
   if (count < sizeof (struct ethhdr)) {
      return -EINVAL;
   }
   
   /*
    * Allocate an sk_buff.
    */
   
   skb = dev_alloc_skb(count + 7);
   if (skb == NULL) {
      // XXX obey O_NONBLOCK?
      return -ENOBUFS;
   }
#ifndef KERNEL_2_1
   skb->free = 1;
#endif
   
   skb_reserve(skb, 2);
   
   /*
    * Copy the data and send it.
    */
   
   userIf->stats.written++;
   memcpy_fromfs(skb_put(skb, count), buf, count);
   
   VNetSend(&userIf->port.jack, skb);

   return count;
}


#ifdef KERNEL_2_1
/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfPoll --
 *
 *      The virtual network's file poll operation.
 *
 * Results: 
 *      Return POLLIN if success, else sleep and return 0.
 *      FIXME: Should not we always return POLLOUT?
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetUserIfPoll(VNetPort     *port,
               struct file  *filp,
	       poll_table   *wait)
{
   VNetUserIF *userIf = (VNetUserIF*)port->jack.private;
   
   poll_wait(filp, &userIf->waitQueue, wait);
   if (!skb_queue_empty(&userIf->packetQueue)) {
      return POLLIN;
   }

   return 0;
}

#else

/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfSelect --
 *
 *      The virtual network's file select operation.
 *
 * Results: 
 *      Return 1 if success, else sleep and return 0.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

static int
VNetUserIfSelect(VNetPort      *port,
                 struct file   *filp,
                 int            sel_type,
                 select_table  *wait)
{
   VNetUserIF *userIf = (VNetUserIF*)port->jack.private;
   
   switch (sel_type) {
   case SEL_IN:
      start_bh_atomic();
      if (!skb_queue_empty(&userIf->packetQueue)) {
	 end_bh_atomic();
	 return 1;
      }
      select_wait(&userIf->waitQueue, wait);
      end_bh_atomic();
      break;
   case SEL_EX:
      // XXX should it be 0 or 1?
      break;
   case SEL_OUT:
      return 1;
   }
   return 0;
}

#endif

/*
 *----------------------------------------------------------------------
 *
 * VNetFileOpIoctl --
 *
 *      The virtual network's ioctl file operation. This is used for
 *      setup of the connection. Currently supported commands are
 *      (taken from sockios.h):
 *
 *      SIOCGIFADDR - get ethernet address          - ioarg OUT: 6 bytes
 *      SIOCSIFADDR - set ethernet address          - ioarg IN:  6 bytes
 *      SIOCGIFFLAGS - get flags                    - ioarg OUT  4 bytes
 *      SIOCSIFFLAGS - set flags                    - ioarg IN:  4 bytes
 *      SIOCGIFBR - get bridging status             - ioarg OUT: 4 bytes
 *      SIOCSIFBR - set bridge peer interface       - ioarg IN:  8 bytes
 *
 *      Private ioctl calls, taken from device-private ioctl space 
 *      in sockios.h, and defined in includes/vm_oui.h:
 *
 *      SIOCSKEEP (0x89F0) - set dealloc flag       - ioarg IN:  1 byte
 *      SIOCGKEEP (0x89F1) - get dealloc flag       - ioarg OUT: 1 byte
 *         By default all hubs (and their associated interfaces) are 
 *         deallocated from memory when their last open port is closed.  
 *         Setting this true "keeps" the hub and thus we will not 
 *         deallocate it until the module is unloaded.
 *
 *      SIOCSLADRF (0x89F2) - set logical address filter (for 
 *         filtering multicast packets)             - ioarg IN:  8 bytes
 *
 *
 *      Supported flags are (taken from if.h):
 *
 *      IFF_UP - ready to receive packets             - OFF by default
 *      IFF_BROADCAST - receive broadcast packets     - OFF by default
 *      IFF_DEBUG - turn on debugging                 - OFF by default
 *      IFF_PROMISC - promiscuous mode                - OFF by default
 *      IFF_MULTICAST - receive multicast packets     - OFF by default
 *      IFF_ALLMULTI - receive all multicast packets 
 *            (like IFF_PROMISC but with multicast)   - OFF by default
 *
 * Results: 
 *      On success 0 else errno.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */

int
VNetUserIfIoctl(VNetPort       *port,
                struct file    *filp,
                unsigned int    iocmd,
                unsigned long   ioarg)
{
   VNetUserIF *userIf = (VNetUserIF*)port->jack.private;

   switch (iocmd) {
   case SIOCSETNOTIFY: { //ORs pollMask into the integer pointed to
                         //by ptr if pending packet. Is cleared
		        	          //when all packets are drained.
      void *pollPtr; 
      if (userIf->pollPtr) {
         return -EINVAL;
      }
      if (verify_area(VERIFY_READ, (void *)ioarg, sizeof(uint32)*2)) {
         return -EFAULT;
      }

      memcpy_fromfs(&pollPtr, (uint32 *)ioarg, sizeof(uint32));
      memcpy_fromfs(&userIf->pollMask, (uint32 *)ioarg+1, sizeof(uint32));

      if (verify_area(VERIFY_WRITE, pollPtr,  sizeof(uint32))) {
         return -EFAULT;
      }

      if (VNetUserIfMapPollPtr(userIf, (uint32)pollPtr)) {
         return -EINVAL;
      }
      userIf->clusterCount = 1;
      break;
   }

   case SIOCUNSETNOTIFY:
      if (!userIf->pollPtr) {
         return -EINVAL;
      }
      VNetUserIfUnmapPollPtr(userIf);
      userIf->clusterCount = 0;
      break;

   case SIOCSETCLUSTERSIZE:
      
      if (!userIf->pollPtr) {
         return -EINVAL;
      }
      userIf->clusterCount = ioarg;

      break;

   default:
      return -ENOIOCTLCMD;
      break;
   }
   
   return 0;
}


/*
 *----------------------------------------------------------------------
 *
 * VNetUserIfProcRead --
 *
 *      Callback for read operation on this userif entry in vnets proc fs.
 *
 * Results: 
 *      Length of read operation.
 *
 * Side effects:
 *      None.
 *
 *----------------------------------------------------------------------
 */
#ifdef KERNEL_2_1
int
VNetUserIfProcRead(char    *page,   // IN/OUT: buffer to write into
                   char   **start,  // OUT: 0 if file < 4k, else offset into page
                   off_t    off,    // IN: offset of read into the file
                   int      count,  // IN: maximum number of bytes to read
                   int     *eof,    // OUT: TRUE if there is nothing more to read
                   void    *data)   // IN: client data - not used
{
   VNetUserIF *userIf = (VNetUserIF*)data; 
   int len = 0;
   
   if (!userIf) {
      return len;
   }
   
   len += VNetPrintPort(&userIf->port, page+len);
   
   len += sprintf(page+len, "read %u written %u queued %u ",
                  userIf->stats.read,
                  userIf->stats.written,
                  userIf->stats.queued);
   
   len += sprintf(page+len, "dropped.down %u dropped.mismatch %u dropped.overflow %u ",
                  userIf->stats.droppedDown,
                  userIf->stats.droppedMismatch,
                  userIf->stats.droppedOverflow);

   len += sprintf(page+len, "\n");
   
   *start = 0;
   *eof   = 1;
   return len;
}
#endif
